/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.cloudstack;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.security.SignatureException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.TreeSet;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.dasein.cloud.AbstractCloud;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.cloudstack.compute.CloudstackComputeServices;
import org.dasein.cloud.cloudstack.network.CloudstackNetworkServices;
import org.dasein.cloud.storage.StorageServices;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class CloudstackProvider extends AbstractCloud {
    static private final Logger logger = Logger.getLogger(CloudstackProvider.class);
    
    static public class ParsedError {
        public int code;
        public String message;
    }
    
    public CloudstackProvider() { }

    public String buildUrl(String command, Param ... params) throws SignatureException {
        return buildUrl(command, false, params);
    }
    
    public String buildUrl(String command, boolean debug, Param ... params) throws SignatureException {
        try {
            StringBuilder str = new StringBuilder();
            String apiKey = new String(getContext().getAccessPublic(), "utf-8");
            String accessKey = new String(getContext().getAccessPrivate(), "utf-8");
            
            StringBuilder newKey = new StringBuilder();
            for( int i =0; i<apiKey.length(); i++ ) {
                char c = apiKey.charAt(i);
                
                if( c != '\r' ) {
                    newKey.append(c);
                }
            }
            apiKey = newKey.toString();
            newKey = new StringBuilder();
            for( int i =0; i<accessKey.length(); i++ ) {
                char c = accessKey.charAt(i);
                
                if( c != '\r' ) {
                    newKey.append(c);
                }
            }
            accessKey = newKey.toString();
            str.append(getHostUrl());
            str.append("/client/api?command=");
            str.append(command);
            for( Param param : params ) {
                str.append("&");
                str.append(param.getKey());
                str.append("=");
                str.append(URLEncoder.encode(param.getValue(), "UTF-8").replaceAll("\\+", "%20"));
            }
            str.append("&apiKey=");
            str.append(URLEncoder.encode(apiKey, "UTF-8").replaceAll("\\+", "%20"));
            str.append("&signature=");
            str.append(URLEncoder.encode(getSignature(command, debug, apiKey, accessKey, params), "UTF-8").replaceAll("\\+", "%20"));
            if( debug ) {
                System.out.println(str.toString());
            }
            return str.toString();
        }
        catch( UnsupportedEncodingException e ) {
            e.printStackTrace();
            throw new RuntimeException("This cannot happen: " + e.getMessage());
        }
    }

    @Override
    public String getCloudName() {
        String name = getContext().getCloudName();
        
        if( name == null ) {
            return "Cloud.com";
        }
        return name;
    }
    
    @Override
    public CloudstackComputeServices getComputeServices() {
        return new CloudstackComputeServices(this);
    }
    
    @Override
    public Zones getDataCenterServices() {
        return new Zones(this);
    }
    
    private String getHostUrl() {
        ProviderContext ctx = getContext();
        
        return ctx.getEndpoint();
    }
    
    @Override
    public CloudstackNetworkServices getNetworkServices() {
        return new CloudstackNetworkServices(this);
    }
    
    @Override
    public String getProviderName() {
        String name = getContext().getProviderName();
        
        if( name == null ) {
            return "Cloud.com";
        }
        return name;
    }

    public CloudstackVersion getVersion() {
        return CloudstackVersion.CS21;
    }
    
    private byte[] calculateHmac(String data, String key) throws SignatureException {
        try {
            SecretKeySpec signingKey = new SecretKeySpec(key.getBytes(), "HmacSHA1");
            Mac mac = Mac.getInstance("HmacSHA1");
            mac.init(signingKey);
        
            return mac.doFinal(data.getBytes());
        } 
        catch (Exception e) {
            throw new SignatureException("Failed to generate HMAC : " + e.getMessage());
        }
    }

    private String getSignature(String command, boolean debug, String apiKey, String accessKey, Param ... params) throws UnsupportedEncodingException, SignatureException {
        TreeSet<Param> sorted = new TreeSet<Param>();
        StringBuilder str = new StringBuilder();
        
        sorted.add(new Param("command", URLEncoder.encode(command, "UTF-8").replaceAll("\\+", "%20").toLowerCase()));
        sorted.add(new Param("apikey", URLEncoder.encode(apiKey, "UTF-8").replaceAll("\\+", "%20").toLowerCase()));
        for( Param param : params ) {
            sorted.add(new Param(param.getKey().toLowerCase(), URLEncoder.encode(param.getValue(), "UTF-8").replaceAll("\\+", "%20").toLowerCase()));
        }
        boolean first = true;
        for( Param param : sorted ) {
            if( !first ) {
                str.append("&");
            }
            first = false;
            str.append(param.getKey());
            str.append("=");
            str.append(param.getValue());
        }
        if( debug ) {
            System.out.println("To sign: " + str.toString());
        }
        return new String(Base64.encodeBase64(calculateHmac(str.toString(), accessKey)));
    }
    
    public ParsedError parseError(int httpStatus, InputStream responseBodyAsStream) throws InternalException {
        ParsedError error = new ParsedError();
        
        error.code = httpStatus;
        error.message = null;
        try {
            Document doc = parseResponse(responseBodyAsStream);
            
            NodeList codes = doc.getElementsByTagName("errorcode");
            for( int i=0; i<codes.getLength(); i++ ) {
                Node n = codes.item(i);
                
                if( n != null && n.hasChildNodes() ) {
                    error.code = Integer.parseInt(n.getFirstChild().getNodeValue().trim());
                }
            }
            NodeList text = doc.getElementsByTagName("errortext");
            for( int i=0; i<text.getLength(); i++ ) {
                Node n = text.item(i);
                
                if( n != null && n.hasChildNodes() ) {
                    error.message = n.getFirstChild().getNodeValue();
                }
            }
        }
        catch( Throwable ignore ) {
            // ignore
        }
        if( error.message == null ) {
            if( httpStatus == 401 ) {
                error.message = "Unauthorized user";
            }
            else if( httpStatus == 430 ) {
                error.message = "Malformed parameters";
            }
            else if( httpStatus == 547 || httpStatus == 530 ) {
                error.message = "Server error in cloud (" + httpStatus + ")";
            }
            else if( httpStatus == 531 ) {
                error.message = "Unable to find account";
            }
            else {
                error.message = "Received error code from server: " + httpStatus;
            }
        }
        return error;
    }
    
    public Document parseResponse(InputStream responseBodyAsStream) throws CloudException, InternalException {
        return parseResponse(responseBodyAsStream, false);
    }
    
    public Document parseResponse(InputStream responseBodyAsStream, boolean debug) throws CloudException, InternalException {
        try {
            BufferedReader in = new BufferedReader(new InputStreamReader(responseBodyAsStream));
            
            StringBuffer sb = new StringBuffer();
            String line;
            
            while ((line = in.readLine()) != null) {
                sb.append(line);
            }
            in.close();
            
            if( debug ) {
                System.out.println(sb.toString());
            }
            ByteArrayInputStream bas = new ByteArrayInputStream(sb.toString().getBytes());
            
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder parser = factory.newDocumentBuilder();
            Document doc = parser.parse(bas);
            bas.close();
            
            return doc;
        }
        catch( IOException e ) {
            throw new CloudException(e);
        }
        catch( ParserConfigurationException e ) {
            throw new CloudException(e);
        }
        catch( SAXException e ) {
            throw new CloudException(e);
        }
    }
    
    public long parseTime(String timestamp) {
        DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ"); //2009-02-03T05:26:32.612278
        
        try {
            return df.parse(timestamp).getTime();
        }
        catch( ParseException e ) {
            df = new SimpleDateFormat("EEE MMM dd HH:mm:ss z yyyy"); //Sun Jul 04 02:18:02 EST 2010
            
            try {
                return df.parse(timestamp).getTime();
            }
            catch( ParseException another ) {
                return 0L;
            }
        }        
    }
    
    @Override
    public String testContext() {
        if( logger.isDebugEnabled() ) {
            logger.debug("enter - " + CloudstackProvider.class.getName() + ".testContext()");
        }
        try {
            try {
                if( logger.isDebugEnabled() ) {
                    logger.debug("testContext(): Checking Cloudstack compute credentials");
                }
                if( !getComputeServices().getVirtualMachineSupport().isSubscribed() ) {
                    logger.warn("testContext(): Cloudstack compute credentials are not subscribed for VM services");
                    return null;
                }
                if( hasStorageServices() ) {
                    if( logger.isDebugEnabled() ) {
                        logger.debug("testContext(): Checking " + getContext().getStorage() + " storage credentials");
                    }
                    StorageServices services = getStorageServices();
                    
                    if( services.hasBlobStoreSupport() ) {
                        try {
                            services.getBlobStoreSupport().listFiles(null).iterator().hasNext();
                        }
                        catch( Throwable t ) {
                            logger.warn("testContext(): Storage credentials failed: " + t.getMessage());
                            if( logger.isDebugEnabled() ) {
                                t.printStackTrace();
                            }
                            return null;
                        }
                    }
                }
                if( logger.isInfoEnabled() ) {
                    logger.info("testContext(): Credentials validated");
                }
                return getContext().getAccountNumber();
            }
            catch( Throwable t ) {
                logger.warn("testContext(): Failed to test cloudstack context: " + t.getMessage());
                if( logger.isDebugEnabled() ) {
                    t.printStackTrace();
                }
                return null;
            }
        }
        finally {
            if( logger.isDebugEnabled() ) {
                logger.debug("exit - " + CloudstackProvider.class.getName() + ".testContext()");
            }
        }
    }
    
    public Document waitForJob(Document doc, String jobName) throws CloudException, InternalException {
        NodeList matches = doc.getElementsByTagName("jobid");
        if( matches.getLength() > 0 ) {
            return waitForJob(matches.item(0).getFirstChild().getNodeValue(), jobName);
        }    
        return null;
    }
    
    public Document waitForJob(String jobId, String jobName) throws CloudException, InternalException {
        while( true ) {
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            HttpClient client = new HttpClient();
            GetMethod get = new GetMethod();
            Document doc;
            int code;
            
            try {
                get.setPath(buildUrl("queryAsyncJobResult", new Param[] { new Param("jobId", jobId) }));
                get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
            }
            catch( SignatureException e ) {
                throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
            }
            try {
                code = client.executeMethod(get);
            }
            catch( HttpException e ) {
                throw new InternalException("HttpException during GET: " + e.getMessage());
            }
            catch( IOException e ) {
                throw new CloudException("IOException during GET: " + e.getMessage());
            }
            try {
                if( code != HttpStatus.SC_OK ) {
                    throw new CloudstackException(parseError(code, get.getResponseBodyAsStream()));
                }
                doc = parseResponse(get.getResponseBodyAsStream());
            }
            catch( IOException e ) {
                throw new CloudException("IOException getting stream: " + e.getMessage());
            } 
            NodeList matches = doc.getElementsByTagName("jobstatus");
            int status = 0;
            
            if( matches.getLength() > 0 ) {
                status = Integer.parseInt(matches.item(0).getFirstChild().getNodeValue());
            }
            if( status > 0 ) {
                if( status == 1 ) {
                    return doc;
                }
                if( status == 2 ) {
                    matches = doc.getElementsByTagName("jobresult");
                    if( matches.getLength() > 0 ) {
                        String str = matches.item(0).getFirstChild().getNodeValue();
                        
                        if( str == null || str.trim().length() < 1 ) {
                            NodeList nodes = matches.item(0).getChildNodes();
                            String message = null;
                            
                            for( int i=0; i<nodes.getLength(); i++ ) {
                                Node n = nodes.item(i);
                                
                                if( n.getNodeName().equalsIgnoreCase("errorcode") ) {
                                    try {
                                        code = Integer.parseInt(n.getFirstChild().getNodeValue().trim());
                                    }
                                    catch( NumberFormatException ignore ) {
                                        // ignore
                                    }
                                }
                                else if( n.getNodeName().equalsIgnoreCase("errortext") ) {
                                    message = n.getFirstChild().getNodeValue().trim();
                                }
                            }
                            ParsedError error = new ParsedError();
                            
                            error.code = code;
                            error.message = message;
                            throw new CloudstackException(error);
                        }
                        else {
                            throw new CloudException(str);
                        }
                    }
                    else {
                        throw new CloudException(jobName + " failed with an unexplained error.");
                    }
                }
            }
        }
    }
}
