/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.cloudstack.compute;

import java.io.IOException;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.cloudstack.CloudstackException;
import org.dasein.cloud.cloudstack.CloudstackProvider;
import org.dasein.cloud.cloudstack.CloudstackProvider.ParsedError;
import org.dasein.cloud.cloudstack.CloudstackVersion;
import org.dasein.cloud.cloudstack.Param;
import org.dasein.cloud.compute.Snapshot;
import org.dasein.cloud.compute.SnapshotState;
import org.dasein.cloud.compute.SnapshotSupport;
import org.dasein.util.CalendarWrapper;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Snapshots implements SnapshotSupport {
    static private final Logger logger = Logger.getLogger(Snapshots.class);
    
    static private final String CREATE_SNAPSHOT = "createSnapshot";
    static private final String DELETE_SNAPSHOT = "deleteSnapshot";
    static private final String LIST_SNAPSHOTS  = "listSnapshots";
    
    private CloudstackProvider provider;
    
    Snapshots(CloudstackProvider provider) {
        this.provider = provider;
    }
    
    @Override
    public String create(String volumeId, String description) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(CREATE_SNAPSHOT, new Param[] { new Param("volumeId", volumeId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                if( code == 431 ) {
                    return null;
                }
                else if( provider.getVersion().equals(CloudstackVersion.CS21) && (code == 500 || code == 530) ) {
                    String msg = get.getResponseBodyAsString();
                    
                    if( msg != null && msg.contains("Snapshot could not be scheduled") ) {
                        // a couple of problems here...
                        // this is not really an error condition, so we should look for the current in-progress snapshot
                        // but cloud.com does not list in-progress snapshots
                        long then = (System.currentTimeMillis() - (CalendarWrapper.MINUTE*9));
                        long now = System.currentTimeMillis() - CalendarWrapper.MINUTE;
                        long timeout = (now + (CalendarWrapper.MINUTE*20));
                        Snapshot wtf = null;
                        
                        while( System.currentTimeMillis() < timeout ) {
                            Snapshot latest = getLatestSnapshot(volumeId);
                            
                            if( latest != null && latest.getSnapshotTimestamp() >= now ) {
                                return latest.getProviderSnapshotId();
                            }
                            else if( latest != null && latest.getSnapshotTimestamp() >= then ) {
                                wtf = latest;
                            }
                            try { Thread.sleep(20000L); }
                            catch( InterruptedException e ) { }
                        }
                        if( wtf != null ) {
                            return wtf.getProviderSnapshotId();
                        }
                        return create(volumeId, description);
                    }
                    else {
                        ParsedError error = new ParsedError();
                        
                        error.code = code;
                        error.message = msg;
                        throw new CloudstackException(error);
                    }
                }
                throw new CloudstackException(provider.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches;
        
        if( provider.getVersion().greaterThan(CloudstackVersion.CS21) ) {
            matches = doc.getElementsByTagName("id");
        }
        else {
            matches = doc.getElementsByTagName("snapshotid");
        }
        String snapshotId = null;
        
        if( matches.getLength() > 0 ) {
            snapshotId = matches.item(0).getFirstChild().getNodeValue();
        }
        if( snapshotId == null ) {
            throw new CloudException("Failed to create a snapshot");
        }
        try {
            provider.waitForJob(doc, "Create Snapshot");
        }
        catch( CloudstackException e ) {
            if( e.getCode() == 431 ) {
                logger.warn("Cloudstack opted not to make a snapshot: " + e.getMessage());
                return null;
            }
            throw e;
        }
        catch( CloudException e ) {
            String msg = e.getMessage();
            
            if( msg != null && msg.contains("no change since last snapshot") ) {
                return null;
            }
            throw e;
        }
        return snapshotId;
    }

    @Override
    public void remove(String snapshotId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(DELETE_SNAPSHOT, new Param[] { new Param("id", snapshotId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(provider.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }        
        provider.waitForJob(doc, "Delete Snapshot");
    }

    @Override
    public String getProviderTermForSnapshot(Locale locale) {
        return "snapshot";
    }

    @Override
    public Iterable<String> listShares(String snapshotId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    @Override
    public Snapshot getSnapshot(String snapshotId) throws InternalException, CloudException {
        if( snapshotId == null ) {
            return null;
        }
        for( Snapshot snapshot : listSnapshots(snapshotId) ) {
            if( snapshot.getProviderSnapshotId().equals(snapshotId) ) {
                return snapshot;
            }
        }
        return null;
    }

    @Override
    public boolean isPublic(String snapshotId) throws InternalException, CloudException {
        return false;
    }

    @Override
    public boolean supportsSnapshotSharing() throws InternalException, CloudException {
        return false;
    }

    @Override
    public boolean supportsSnapshotSharingWithPublic() throws InternalException, CloudException {
        return false;
    }

    @Override
    public Iterable<Snapshot> listSnapshots() throws InternalException, CloudException {
        return listSnapshots(null);
    }
    
    private Collection<Snapshot> listSnapshots(String snapshotId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_SNAPSHOTS, new Param[] { new Param("zoneId", provider.getContext().getRegionId()) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(provider.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<Snapshot> snapshots = new ArrayList<Snapshot>();
        NodeList matches = doc.getElementsByTagName("snapshot");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node s = matches.item(i);

            if( s != null ) {
                Snapshot snapshot = toSnapshot(s);
                
                if( snapshot != null ) {
                    snapshots.add(snapshot);
                }
            }
        }
        return snapshots;
    }

    private Snapshot getLatestSnapshot(String forVolumeId) throws InternalException, CloudException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(provider.buildUrl(LIST_SNAPSHOTS, new Param[] { new Param("zoneId", provider.getContext().getRegionId()), new Param("volumeId", forVolumeId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(provider.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = provider.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        Snapshot latest = null;
        
        NodeList matches = doc.getElementsByTagName("snapshot");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node s = matches.item(i);

            if( s != null ) {
                Snapshot snapshot = toSnapshot(s);
                
                if( snapshot != null && snapshot.getVolumeId() != null && snapshot.getVolumeId().equals(forVolumeId) ) {
                    if( latest == null || snapshot.getSnapshotTimestamp() > latest.getSnapshotTimestamp() ) {
                        latest = snapshot;
                    }
                }
            }
        }
        return latest;
    }
    
    @Override
    public void shareSnapshot(String arg0, String arg1, boolean arg2) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    private Snapshot toSnapshot(Node node) {
        Snapshot snapshot = new Snapshot();
        NodeList attrs = node.getChildNodes();
        String type = null;
        
        snapshot.setCurrentState(SnapshotState.AVAILABLE);
        snapshot.setOwner(provider.getContext().getAccountNumber());
        snapshot.setProgress("100%");
        snapshot.setRegionId(provider.getContext().getRegionId());
        snapshot.setSizeInGb(0);
        for( int i=0; i<attrs.getLength(); i++ ) {
            Node attr = attrs.item(i);
            
            String name = attr.getNodeName();
            String value = null;
            
            if( attr.hasChildNodes() ) {
                value = attr.getFirstChild().getNodeValue();
            }
            if( name.equalsIgnoreCase("id") ) {
                snapshot.setProviderSnapshotId(value);
            }
            else if( name.equalsIgnoreCase("volumeid") ) {
                snapshot.setVolumeId(value);
            }
            else if( name.equalsIgnoreCase("name") ) {
                snapshot.setName(value);
            }
            else if( name.equalsIgnoreCase("created") ) {
                snapshot.setSnapshotTimestamp(provider.parseTime(value));
            }
            else if( name.equalsIgnoreCase("snapshottype") ) {
                type = value;
            }
            else if( name.equals("account") ) {
                snapshot.setOwner(value);
            }
        } 
        if( snapshot.getProviderSnapshotId() == null ) {
            return null;
        }
        if( snapshot.getName() == null ) {
            snapshot.setName(snapshot.getProviderSnapshotId());
        }
        if( snapshot.getDescription() == null ) {
            if( type == null ) {
                snapshot.setDescription(snapshot.getName());
            }
            else {
                snapshot.setDescription(snapshot.getName() + " (" + type + ")");            
            }
        }
        return snapshot;
    }

    @Override
    public boolean isSubscribed() throws InternalException, CloudException {
        return provider.getComputeServices().getVolumeSupport().isSubscribed();
    }
}
