/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.cloudstack.network;

import java.io.IOException;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Locale;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.cookie.CookiePolicy;
import org.apache.commons.httpclient.methods.GetMethod;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.cloudstack.CloudstackException;
import org.dasein.cloud.cloudstack.CloudstackProvider;
import org.dasein.cloud.cloudstack.Param;
import org.dasein.cloud.cloudstack.Zones;
import org.dasein.cloud.network.NetworkInterface;
import org.dasein.cloud.network.Subnet;
import org.dasein.cloud.network.VLANSupport;
import org.dasein.cloud.network.VLAN;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class Network implements VLANSupport {
    static public final String CREATE_NETWORK         = "createNetwork";
    static public final String LIST_NETWORK_OFFERINGS = "listNetworkOfferings";
    static public final String LIST_NETWORKS          = "listNetworks";
    
    private CloudstackProvider cloudstack;
    
    Network(CloudstackProvider cloudstack) { this.cloudstack = cloudstack; }

    @Override
    public boolean allowsNewSubnetCreation() throws CloudException, InternalException {
        return false;
    }
    
    @Override
    public boolean allowsNewVlanCreation() throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(cloudstack.buildUrl(Zones.LIST_ZONES, new Param[] { new Param("id", cloudstack.getContext().getRegionId()) }));
            get.addRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=utf-8");
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            return false;
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            return false;
        }
        catch( IOException e ) {
            return false;
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                return false;
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( Throwable t ) {
            return false;
        }
        NodeList matches = doc.getElementsByTagName("zone");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            NodeList attrs = node.getChildNodes();
            
            for( int j=0; j<attrs.getLength(); j++ ) {
                Node attr = attrs.item(j);
                                                         
                if( attr.getNodeName().equalsIgnoreCase("securitygroupsenabled") ) {
                    String val = null;
                    
                    if( attr.hasChildNodes() ) {
                        val = attr.getFirstChild().getNodeValue().trim();
                    }
                    if( val != null && val.equalsIgnoreCase("true") ) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    @Override
    public int getMaxVlanCount() throws CloudException, InternalException {
        return 1;
    }

    private String getNetworkOffering(String inZoneId) throws InternalException, CloudException {
        String type = "virtual";
        
        {
            HttpClient client = new HttpClient();
            GetMethod get = new GetMethod();
            Document doc;
            int code;
            
            try {
                get.setPath(cloudstack.buildUrl(Zones.LIST_ZONES, new Param[] { new Param("id", cloudstack.getContext().getRegionId()) }));
                get.addRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=utf-8");
                get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
            }
            catch( SignatureException e ) {
                throw new InternalException(e);
            }
            try {
                code = client.executeMethod(get);
            }
            catch( HttpException e ) {
                throw new CloudException(e);
            }
            catch( IOException e ) {
                throw new CloudException(e);
            }
            try {
                if( code == HttpStatus.SC_OK ) {
                    doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
                }
                else {
                    doc = null;
                }
            }
            catch( Throwable t ) {
                throw new InternalException(t);
            }
            NodeList matches = doc.getElementsByTagName("zone");
            for( int i=0; i<matches.getLength(); i++ ) {
                Node node = matches.item(i);
                NodeList attrs = node.getChildNodes();
                
                for( int j=0; j<attrs.getLength(); j++ ) {
                    Node attr = attrs.item(j);
                                                             
                    if( attr.getNodeName().equalsIgnoreCase("securitygroupsenabled") ) {
                        String val = null;
                        
                        if( attr.hasChildNodes() ) {
                            val = attr.getFirstChild().getNodeValue().trim();
                        }
                        if( val != null && val.equalsIgnoreCase("true") ) {
                            type = "direct";
                        }
                    }
                }
            }
        }
        
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
            
        try {
            get.setPath(cloudstack.buildUrl(LIST_NETWORK_OFFERINGS, new Param[] { new Param("zoneId", inZoneId) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(cloudstack.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("networkoffering");
        String offeringId = null;
        
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            NodeList attributes;
            boolean available = false, virtual = false;
            String id = null;
            
            attributes = node.getChildNodes();
            for( int j=0; j<attributes.getLength(); j++ ) {
                Node n = attributes.item(j);
                
                String value;

                if( n.getChildNodes().getLength() > 0 ) {
                    value = n.getFirstChild().getNodeValue();
                }
                else {
                    value = null;
                }
                if( n.getNodeName().equals("id") ) {
                    id = value;
                }
                else if( n.getNodeName().equalsIgnoreCase("availability") ) {
                    available = (value != null && !value.equalsIgnoreCase("unavailable"));
                }
                else if( n.getNodeName().equalsIgnoreCase("guestiptype") ) {
                    virtual = (value != null && value.equalsIgnoreCase(type));                    
                }
            }
            if( available && virtual) {
                offeringId = id;
                break;
            }
        }
        return offeringId;
    }
    
    @Override
    public VLAN getVlan(String vlanId) throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(cloudstack.buildUrl(Network.LIST_NETWORKS, new Param("zoneId", cloudstack.getContext().getRegionId()), new Param("id", vlanId)));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code == 431 ) {
                return null;
            }
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(cloudstack.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("network");
        
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
                
            if( node != null ) {
                VLAN vlan = toNetwork(node);
                    
                if( vlan != null ) {
                    return vlan;
                }
            }
        }
        return null;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(cloudstack.buildUrl(Zones.LIST_ZONES, new Param[] { new Param("id", cloudstack.getContext().getRegionId()) }));
            get.addRequestHeader("Content-Type", "application/x-www-form-urlencoded; charset=utf-8");
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(cloudstack.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        NodeList matches = doc.getElementsByTagName("zone");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            NodeList attrs = node.getChildNodes();
            
            for( int j=0; j<attrs.getLength(); j++ ) {
                Node attr = attrs.item(j);
                
                if( attr.getNodeName().equalsIgnoreCase("networktype") ) {
                    String val = null;
                    
                    if( attr.hasChildNodes() ) {
                        val = attr.getFirstChild().getNodeValue().trim();
                    }
                    if( val == null || val.equalsIgnoreCase("Basic") ) {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    @Override
    public Iterable<VLAN> listVlans() throws CloudException, InternalException {
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            get.setPath(cloudstack.buildUrl(Network.LIST_NETWORKS, new Param("zoneId", cloudstack.getContext().getRegionId()),  new Param("isshared", "true")));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Unable to generate a valid signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(cloudstack.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }
        ArrayList<VLAN> networks = new ArrayList<VLAN>();
        NodeList matches = doc.getElementsByTagName("network");
        
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
                
            if( node != null ) {
                VLAN vlan = toNetwork(node);
                    
                if( vlan != null ) {
                    networks.add(vlan);
                }
            }
        }
        return networks;
    }

    @Override
    public VLAN createVlan(String cidr, String name, String description, String domainName, String[] dnsServers, String[] ntpServers) throws CloudException, InternalException {
        if( !allowsNewVlanCreation() ) {
            throw new OperationNotSupportedException();
        }
        HttpClient client = new HttpClient();
        GetMethod get = new GetMethod();
        Document doc;
        int code;
        
        try {
            String offering = getNetworkOffering(cloudstack.getContext().getRegionId());
            
            // TODO: specify subnet
            get.setPath(cloudstack.buildUrl(CREATE_NETWORK, new Param[] { new Param("zoneId", cloudstack.getContext().getRegionId()), new Param("networkOfferingId", offering), new Param("name", name), new Param("displayText", name) }));
            get.getParams().setCookiePolicy(CookiePolicy.IGNORE_COOKIES);
        }
        catch( SignatureException e ) {
            throw new InternalException("Failed to generate a proper signature: " + e.getMessage());
        }
        try {
            code = client.executeMethod(get);
        }
        catch( HttpException e ) {
            throw new InternalException("HttpException during GET: " + e.getMessage());
        }
        catch( IOException e ) {
            throw new CloudException("IOException during GET: " + e.getMessage());
        }
        try {
            if( code != HttpStatus.SC_OK ) {
                throw new CloudstackException(cloudstack.parseError(code, get.getResponseBodyAsStream()));
            }
            doc = cloudstack.parseResponse(get.getResponseBodyAsStream());
        }
        catch( IOException e ) {
            throw new CloudException("IOException getting stream: " + e.getMessage());
        }    
        NodeList matches = doc.getElementsByTagName("network");
        for( int i=0; i<matches.getLength(); i++ ) {
            Node node = matches.item(i);
            
            if( node != null ) {
                VLAN network = toNetwork(node);
                
                if( network != null ) {
                    return network;
                }
            }
        }
        throw new CloudException("Creation requested failed to create a network without an error");
    }

    @Override
    public void removeVlan(String vlanId) throws CloudException, InternalException {
        // TODO Auto-generated method stub
        
    }

    public VLAN toNetwork(Node node) {
        if( node == null ) {
            return null;
        }
        VLAN network = new VLAN();
        
        NodeList attributes = node.getChildNodes();

        network.setProviderOwnerId(cloudstack.getContext().getAccountNumber());
        network.setProviderRegionId(cloudstack.getContext().getRegionId());
        
        for( int i=0; i<attributes.getLength(); i++ ) {
            Node attribute = attributes.item(i);
            String name = attribute.getNodeName().toLowerCase();
            String value;
            
            if( attribute.getChildNodes().getLength() > 0 ) {
                value = attribute.getFirstChild().getNodeValue();                
            }
            else {
                value = null;
            }
            if( name.equalsIgnoreCase("id") ) {
                network.setProviderVlanId(value);
            }
            else if( name.equalsIgnoreCase("name") ) {
                if( network.getName() == null ) {
                    network.setName(value);
                }
            }
            else if( name.equalsIgnoreCase("displaytext") ) {
                network.setName(value);
            }
            else if( name.equalsIgnoreCase("zoneid") && value != null ) {
                network.setProviderRegionId(value);
            }
            else if( name.startsWith("dns") && value != null && !value.trim().equals("") ) {
                String[] dns;
                
                if( network.getDnsServers() != null ) {
                    dns = new String[network.getDnsServers().length+1];
                    for( int idx=1; idx<=network.getDnsServers().length; idx++ ) {
                        dns[idx] = network.getDnsServers()[idx-1];
                    }
                }
                else {
                    dns = new String[] { value };
                }
                network.setDnsServers(dns);
            }
        }
        if( network.getProviderVlanId() == null ) {
            return null;
        }
        network.setProviderDataCenterId(network.getProviderRegionId());
        if( network.getName() == null ) {
            network.setName(network.getProviderVlanId());
        }
        if( network.getDescription() == null ) {
            network.setDescription(network.getName());
        }
        return network;
    }

    @Override
    public Iterable<NetworkInterface> listNetworkInterfaces(String arg0) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public Subnet createSubnet(String cidr, String inProviderVlanId, String name, String description) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Unable to create subnets");
    }

    @Override
    public String getProviderTermForNetworkInterface(Locale locale) {
        return "NIC";
    }

    @Override
    public String getProviderTermForSubnet(Locale locale) {
        return "network";
    }

    @Override
    public String getProviderTermForVlan(Locale locale) {
        return "network";
    }

    @Override
    public Subnet getSubnet(String subnetId) throws CloudException, InternalException {
        return null;
    }

    @Override
    public boolean isVlanDataCenterConstrained() throws CloudException, InternalException {
        return true;
    }

    @Override
    public boolean isSubnetDataCenterConstrained() throws CloudException, InternalException {
        return true;
    }
    
    @Override
    public Iterable<Subnet> listSubnets(String inVlanId) throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public void removeSubnet(String providerSubnetId) throws CloudException, InternalException {
        throw new OperationNotSupportedException();
    }

    @Override
    public boolean supportsVlansWithSubnets() throws CloudException, InternalException {
        return false;
    }
}
